import discord 
from discord.ext import commands
import sys
import typing
import pkg_resources
from ..core.work.types import ContextA
from ..core.codeblocks import Codeblock, codeblock_converter
import traceback
import math
import itertools
import re
import time
from urllib.parse import urlencode
import asyncio

from jejudo.modules import ExtensionConverter
from jejudo.shell import ShellReader
from jejudo.exception_handling import ReplResponseReactor
from jejudo.flags import Flags
from jejudo.paginators import PaginatorInterface, WrappedPaginator





# ==============================================================================

try:
    import psutil
except ImportError:
    psutil = None

try:
    from importlib.metadata import distribution, packages_distributions
except ImportError:
    from importlib_metadata import distribution, packages_distributions

# ==============================================================================





def natural_size(size_in_bytes: int) -> str:
    """
    Converts a number of bytes to an appropriately-scaled unit
    E.g.:
        1024 -> 1.00 KiB
        12345678 -> 11.77 MiB
    """
    units = ('B', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB', 'YiB')

    power = int(math.log(max(abs(size_in_bytes), 1), 1024))

    return f"{size_in_bytes / (1024 ** power):.2f} {units[power]}"


def natural_time(time_in_seconds: float) -> str:
    """
    Converts a time in seconds to a 6-padded scaled unit
    E.g.:
        1.5000 ->   1.50  s
        0.1000 -> 100.00 ms
        0.0001 -> 100.00 us
    """
    units = (
        ('mi', 60),
        (' s', 1),
        ('ms', 1e-3),
        ('\N{GREEK SMALL LETTER MU}s', 1e-6),
    )

    absolute = abs(time_in_seconds)

    for label, size in units:
        if absolute > size:
            return f"{time_in_seconds / size:6.2f} {label}"

    return f"{time_in_seconds / 1e-9:6.2f} ns"


def mean_stddev(collection: typing.Collection[float]) -> typing.Tuple[float, float]:
    """
    Takes a collection of floats and returns (mean, stddev) as a tuple.
    """

    average = sum(collection) / len(collection)

    if len(collection) > 1:
        stddev = math.sqrt(sum(math.pow(reading - average, 2) for reading in collection) / (len(collection) - 1))
    else:
        stddev = 0.0

    return (average, stddev)


def format_stddev(collection: typing.Collection[float]) -> str:
    """
    Takes a collection of floats and produces a mean (+ stddev, if multiple values exist) string.
    """
    if len(collection) > 1:
        average, stddev = mean_stddev(collection)

        return f"{natural_time(average)} \N{PLUS-MINUS SIGN} {natural_time(stddev)}"

    return natural_time(sum(collection) / len(collection))


class Dropdown(discord.ui.Select):
    def __init__(self, bot_: discord.Bot, requester):
        self.bot = bot_
        self.requester = requester
        options = [
            discord.SelectOption(
                label="jejudo page", description="View the Jeju Island main page.", emoji="🟥"
            ),
            discord.SelectOption(
                label="protected_access page", description="The page related to the intent.", emoji="🟩"
            )
        ]

        super().__init__(
            placeholder="jejudo !! Click here to open the menu!",
            min_values=1,
            max_values=1,
            options=options,
        )

    async def callback(self, interaction: discord.Interaction):
        if interaction.user != self.requester:
            return await interaction.response.send_message("❌ jejudo is the commands owner !!", ephemeral=True)

        if self.values[0] == "jejudo page":

            jejudo = [
                f"> the, jejudo ` v{pkg_resources.get_distribution('jejudo').version} `, py-cord ` v{pkg_resources.get_distribution('py-cord').version} `".replace("\n", ""),
                f"ㄴ • Python ` {sys.version}`".replace("\n", ""),
                f"ㄴ • platform `{sys.platform}`".replace("\n", "")
            ]
            
            if psutil:
                try:
                    proc = psutil.Process()

                    with proc.oneshot():
                        try:
                            mem = proc.memory_full_info()
                            jejudo.append(f"\n • {natural_size(mem.rss)} memory , {natural_size(mem.vms)} memory , {natural_size(mem.uss)} memory")
                        except psutil.AccessDenied:
                            pass

                        try:
                            name = proc.name()
                            pid = proc.pid
                            thread_count = proc.num_threads()

                            jejudo.append(f"ㄴ • PID {pid} (`{name}`) == {thread_count} thread(s)")
                        except psutil.AccessDenied:
                            pass

                        jejudo.append("")
                except psutil.AccessDenied:
                    jejudo.append(
                        "ㄴ • psutil is installed, but this process does not have high enough access rights "
                    )
                    jejudo.append("")

            cache_summary = f"` 🌐 {len(self.bot.guilds)} ` guild(s) ㅣ ` 🤴 {len(self.bot.users)} ` user(s)"

            if isinstance(self.bot, discord.AutoShardedClient):
                if len(self.bot.shards) > 20:
                    jejudo.append(
                        f"This bot is auto shards (` {len(self.bot.shards)} ` shards of ` {self.bot.shard_count} `)"
                        f"\n{cache_summary}."
                    )
                else:
                    shard_ids = ', '.join(str(i) for i in self.bot.shards.keys())
                    jejudo.append(
                        f"This bot is auto shards (Shards ` {shard_ids} ` of ` {self.bot.shard_count} `)"
                        f"\n{cache_summary}."
                    )
            elif self.bot.shard_count:
                jejudo.append(
                    f"This bot is manually sharded (Shard ` {self.bot.shard_id} ` of ` {self.bot.shard_count} `)"
                    f"\n{cache_summary}."
                )
            else:
                jejudo.append(f"This bot is not sharded\n{cache_summary}.")

            jejudo.append(f"\n\nAverage websocket latency: {round(self.bot.latency * 1000, 2)}ms")

            await interaction.response.edit_message(
                content="\n".join(jejudo)
            )
        if self.values[0] == "protected_access page":
            jejudo = [
                f"> the, jejudo ` v{pkg_resources.get_distribution('jejudo').version} `, py-cord ` v{pkg_resources.get_distribution('py-cord').version} `".replace("\n", ""),
                f"ㄴ • Python ` {sys.version}`".replace("\n", ""),
                f"ㄴ • platform `{sys.platform}`".replace("\n", "")
            ]

            if self.bot._connection.max_messages:
                message_cache = f"Message cache capped at {self.bot._connection.max_messages}"
            else:
                message_cache = "Message cache is disabled"

            if discord.version_info >= (1, 5, 0):
                remarks = {
                    True: 'enabled',
                    False: 'disabled',
                    None: 'unknown'
                }

                *group, last = (
                    f"{intent.replace('_', ' ')} intent is {remarks.get(getattr(self.bot.intents, intent, None))}\n"
                    for intent in
                    ('presences', 'members', 'message_content')
                )

                jejudo.append(f"\n{message_cache}\n{''.join(group)}\n{last}")
            else:
                guild_subscriptions = f"guild subscriptions are {'enabled' if self.bot._connection.guild_subscriptions else 'disabled'}"  # type: ignore

                jejudo.append(f"\n{message_cache}\n{guild_subscriptions}")
            jejudo.append(f"Average websocket latency: {round(self.bot.latency * 1000, 2)}ms")

            await interaction.response.edit_message(
                content="\n".join(jejudo)
            )

class DropdownView(discord.ui.View):
    def __init__(self, bot_: discord.Bot, requester):
        self.bot = bot_
        self.requester = requester
        super().__init__()
        self.add_item(Dropdown(self.bot, self.requester))

class Jejudos(commands.Cog):
    def __init__(self, bot):
        self.bot: commands.Bot = bot


    @commands.Cog.listener()
    async def on_command_error(self, ctx, error):
        if isinstance(error, commands.NotOwner):
            await ctx.reply("❌ jejudo is the commands owner !!")


    @commands.group("jejudo",aliases=["jeju","jds"],invoke_without_command=True)
    @commands.is_owner()
    async def jejudo_jsk(self, ctx: ContextA):
        """
        The jejudo debug and diagnostic commands.
        This command on its own gives a status brief.
        All other functionality is within its subcommands.
        """
        try:

            jejudo = [
                f"> the, jejudo ` v{pkg_resources.get_distribution('jejudo').version} `, py-cord ` v{pkg_resources.get_distribution('py-cord').version} `".replace("\n", ""),
                f"ㄴ • Python ` {sys.version}`".replace("\n", ""),
                f"ㄴ • platform `{sys.platform}`".replace("\n", "")
            ]
            
            if psutil:
                try:
                    proc = psutil.Process()

                    with proc.oneshot():
                        try:
                            mem = proc.memory_full_info()
                            jejudo.append(f"\n • {natural_size(mem.rss)} memory , {natural_size(mem.vms)} memory , {natural_size(mem.uss)} memory")
                        except psutil.AccessDenied:
                            pass

                        try:
                            name = proc.name()
                            pid = proc.pid
                            thread_count = proc.num_threads()

                            jejudo.append(f"ㄴ • PID {pid} (`{name}`) == {thread_count} thread(s)")
                        except psutil.AccessDenied:
                            pass

                        jejudo.append("")  # blank line
                except psutil.AccessDenied:
                    jejudo.append(
                        "ㄴ • psutil is installed, but this process does not have high enough access rights "
                    )
                    jejudo.append("")  # blank line

            cache_summary = f"` 🌐 {len(self.bot.guilds)} ` guild(s) ㅣ ` 🤴 {len(self.bot.users)} ` user(s)"

            # Show shard settings to summary
            if isinstance(self.bot, discord.AutoShardedClient):
                if len(self.bot.shards) > 20:
                    jejudo.append(
                        f"This bot is auto shards (` {len(self.bot.shards)} ` shards of ` {self.bot.shard_count} `)"
                        f"\n{cache_summary}."
                    )
                else:
                    shard_ids = ', '.join(str(i) for i in self.bot.shards.keys())
                    jejudo.append(
                        f"This bot is auto shards (Shards ` {shard_ids} ` of ` {self.bot.shard_count} `)"
                        f"\n{cache_summary}."
                    )
            elif self.bot.shard_count:
                jejudo.append(
                    f"This bot is manually sharded (Shard ` {self.bot.shard_id} ` of ` {self.bot.shard_count} `)"
                    f"\n{cache_summary}."
                )
            else:
                jejudo.append(f"This bot is not sharded\n{cache_summary}.")
            jejudo.append(f"\n\nAverage websocket latency: {round(self.bot.latency * 1000, 2)}ms")

            view = DropdownView(self.bot, ctx.author)
            await ctx.reply("\n".join(jejudo), view=view)
        except:
            print(traceback.format_exc())




    @jejudo_jsk.command(name="shell", aliases=["bash", "sh", "powershell", "ps1", "ps", "cmd", "terminal"])
    @commands.is_owner()
    async def jejudo_shell(self, ctx: ContextA, *, argument: codeblock_converter=None):
        """
        Run a shell command
        """
        try:
            if not argument:
                await ctx.send("Usage: `jejudo shell <argument>`")
                return
            if typing.TYPE_CHECKING:
                argument: Codeblock = argument  # type: ignore

            try:
                async with ReplResponseReactor(ctx.message):
                    with ShellReader(argument.content) as reader:
                        prefix = "```" + reader.highlight

                        paginator = WrappedPaginator(prefix=prefix, max_size=1975)
                        paginator.add_line(f"{reader.ps1} {argument.content}\n")

                        interface = PaginatorInterface(ctx.bot, paginator, owner=ctx.author)
                        self.bot.loop.create_task(interface.send_to(ctx))

                        async for line in reader:
                            if interface.closed:
                                return
                            await interface.add_line(line)

                    await interface.add_line(f"\n[ jeju status ] Return code {reader.close_code}")
            except:
                print(traceback.format_exc())
        except asyncio.TimeoutError:
            pass



    @jejudo_jsk.command(name="shutdown", aliases=["logout", "out", "doun"])
    @commands.is_owner()
    async def jejudo_shutdown(self, ctx: ContextA):
        """
        Logs this bot out.
        """
        try:

            await ctx.send(f"__**[ ` jejudo ` ]** •  Shutdown now . . .__")
            await ctx.bot.close()
        except:
            print(traceback.format_exc())




    @jejudo_jsk.command(name="rtt", aliases=["ping"])
    @commands.is_owner()
    async def jejudo_rtt(self, ctx: ContextA):
        """
        Calculates Round-Trip Time to the API.
        """
        try:
            message = None

            # We'll show each of these readings as well as an average and standard deviation.
            api_readings: typing.List[float] = []
            # We'll also record websocket readings, but we'll only provide the average.
            websocket_readings: typing.List[float] = []

            # We do 6 iterations here.
            # This gives us 5 visible readings, because a request can't include the stats for itself.
            for _ in range(6):
                # First generate the text
                text = "Calculating round-trip time...\n\n"
                text += "\n".join(f"Reading {index + 1}: {reading * 1000:.2f}ms" for index, reading in enumerate(api_readings))

                if api_readings:
                    average, stddev = mean_stddev(api_readings)

                    text += f"\n\nAverage: {average * 1000:.2f} \N{PLUS-MINUS SIGN} {stddev * 1000:.2f}ms"
                else:
                    text += "\n\nNo readings yet."

                if websocket_readings:
                    average = sum(websocket_readings) / len(websocket_readings)

                    text += f"\nWebsocket latency: {average * 1000:.2f}ms"
                else:
                    text += f"\nWebsocket latency: {self.bot.latency * 1000:.2f}ms"

                # Now do the actual request and reading
                if message:
                    before = time.perf_counter()
                    await message.edit(content=text)
                    after = time.perf_counter()

                    api_readings.append(after - before)
                else:
                    before = time.perf_counter()
                    message = await ctx.send(content=text)
                    after = time.perf_counter()

                    api_readings.append(after - before)

                # Ignore websocket latencies that are 0 or negative because they usually mean we've got bad heartbeats
                if self.bot.latency > 0.0:
                    websocket_readings.append(self.bot.latency)
        except:
            print(traceback.format_exc())





    @jejudo_jsk.command(name="load", aliases=["reload"])
    @commands.is_owner()
    async def jejudo_load(self, ctx: ContextA, *extensions: ExtensionConverter):  # type: ignore
        """
        Loads or reloads the given extension names.
        Reports any extensions that failed to load.
        """
        try:
            extensions: typing.Iterable[typing.List[str]] = extensions  # type: ignore

            paginator = commands.Paginator(prefix='', suffix='')

            # 'jsk reload' on its own just reloads jishaku
            if ctx.invoked_with == 'reload' and not extensions:
                extensions = [['jejudo']]

            for extension in itertools.chain(*extensions):
                method, icon = (
                    (self.bot.reload_extension, "\N{CLOCKWISE RIGHTWARDS AND LEFTWARDS OPEN CIRCLE ARROWS}")
                    if extension in self.bot.extensions else
                    (self.bot.load_extension, "\N{INBOX TRAY}")
                )

                try:
                    await discord.utils.maybe_coroutine(method, extension)
                except Exception as exc:  # pylint: disable=broad-except
                    if isinstance(exc, commands.ExtensionFailed) and exc.__cause__:
                        cause = exc.__cause__
                        traceback_data = ''.join(traceback.format_exception(type(cause), cause, cause.__traceback__, 8))
                    else:
                        traceback_data = ''.join(traceback.format_exception(type(exc), exc, exc.__traceback__, 2))

                    paginator.add_line(
                        f"{icon}\N{WARNING SIGN} `{extension}`\n```py\n{traceback_data}\n```",
                        empty=True
                    )
                else:
                    paginator.add_line(f"{icon} `{extension}`", empty=True)

            for page in paginator.pages:
                await ctx.send(page)
        except:
            print(traceback.format_exc())
